@@ -1,5 +1,31 @@
 #include "utils.h"
 
+// return the current value of a pairlist
+SEXP pairlist_car(SEXP x) {
+  if (!Rf_isList(x))
+    Rf_error("x must be a pairlist");
+  return CAR(x);
+}
+
+// return the next cons of a pairlist
+SEXP pairlist_cdr(SEXP x) {
+  if (!Rf_isList(x))
+    Rf_error("x must be a pairlist");
+  return CDR(x);
+}
+
+// return the last cons of a pairlist
+SEXP pairlist_last(SEXP x) {
+  if (!Rf_isList(x))
+    Rf_error("x must be a pairlist");
+  SEXP nx = CDR(x);
+  while (!Rf_isNull(nx)) {
+      x = nx;
+      nx = CDR(x);
+  }
+  return x;
+}
+
 
 SEXP get_sexp_value(SEXP env, const char* name) {
     SEXP x = Rf_findVarInFrame(env, Rf_install(name));
